import os

# DJANGO ######################################################################

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

DEBUG = True
EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)

ROOT_URLCONF = 'master_password.urls'
SECRET_KEY = 'secret-key'
STATIC_URL = '/static/'

# MASTER PASSWORD #############################################################

AUTHENTICATION_BACKENDS = ('master_password.auth.ModelBackend', )

INSTALLED_APPS += (
    'master_password',
    'master_password.tests',
)

# NOSE ########################################################################

INSTALLED_APPS += ('django_nose', )
TEST_RUNNER = 'django_nose.NoseTestSuiteRunner'  # Default: django.test.runner.DiscoverRunner

NOSE_ARGS = [
    '--with-progressive',  # See https://github.com/erikrose/nose-progressive
]

# TRAVIS ######################################################################

if 'TRAVIS' in os.environ:

    # Disable progressive plugin, which doesn't work properly on Travis.
    NOSE_ARGS.remove('--with-progressive')



import string
import warnings

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.hashers import check_password, make_password
from django.utils.module_loading import import_string
from django.utils.translation import ugettext as _


class MasterPasswordMixin(object):
    """
    Adds fallback master password authentication to an existing backend.

    On each subclass, you must define:

        get_user_object(**kwargs)
            Return a user object to be validated. This will usually be the same
            as ``authenticate()`` on the super class, but without any password
            validation.

    You can also optionally override:

        get_master_passwords()
            Return a dictionary with clear text or hashed passwords as keys,
            and callback functions (or ``None``) as values.

            The callback function for each master password must take a user
            object as its only argument, and should return ``True`` if the user
            is allowed to authenticate with that password.

            The ``MASTER_PASSWORDS`` setting is returned by default.

        get_password(**kwargs)
            Return a password to be validated. The ``password`` keyword
            argument is returned as-is by default.
    """

    def authenticate(self, request, **kwargs):
        """
        Return a user object if authenticated by the superclass or a valid
        master password.
        """
        # First, try to authenticate with the superclass.
        user = super(MasterPasswordMixin, self).authenticate(request, **kwargs)
        if user is not None:
            return user
        # Fallback to master password authentication.
        user = self.get_user_object(**kwargs)
        password = self.get_password(**kwargs)
        if user and password:
            # Try all the master passwords.
            for master, callback in self.get_master_passwords().items():
                if settings.DEBUG:
                    # Check hashed and clear text versions.
                    hashed = [master, make_password(master)]
                else:
                    # Check only hashed version.
                    hashed = [master]
                for master in hashed:
                    if check_password(password, master):
                        if callback is None or callback(user):
                            # Master password *must* be strong in production.
                            if not settings.DEBUG and \
                                    not self.is_strong_password(password):
                                warnings.warn(message=_(
                                    'When DEBUG=False, master passwords must '
                                    'be more than 50 characters, with at '
                                    'least 1 uppercase, 1 lowercase, 1 digit '
                                    'and 1 non-alphanumeric character.'))
                                continue
                            return user

    def get_master_passwords(self):
        """
        Return a dictionary of master passwords.
        """
        return getattr(settings, 'MASTER_PASSWORDS', {})

    def get_password(self, password=None, **kwargs):
        """
        Return the password to be validated from the given keyword arguments.
        """
        return password

    def get_user_object(self, **kwargs):
        """
        Return a user object for the given keyword arguments.
        """
        raise NotImplemented  # pragma: no cover

    def is_strong_password(self, password):
        """
        Return `True` if the password is strong.
        """
        chars = set(password)
        is_strong = all([
            chars.intersection(string.digits),
            chars.intersection(string.ascii_lowercase),
            chars.intersection(string.ascii_uppercase),
            len(chars) >= 50,
            not password.isalnum(),
        ])
        return is_strong


class ModelBackend(MasterPasswordMixin, ModelBackend):
    """
    Add master password authentication to ``ModelBackend``.
    """

    def get_user_object(self, username=None, password=None, **kwargs):
        UserModel = get_user_model()
        if username is None:
            username = kwargs.get(UserModel.USERNAME_FIELD)
        try:
            user = UserModel._default_manager.get_by_natural_key(username)
            return user
        except UserModel.DoesNotExist:
            return None


PRODUCTION_WARNING = (
    'django-master-password is enabled and DEBUG=False. You *must* use '
    'strong hashed master passwords.')


def production_warning():
    # Issue a warning if enabled in production.
    for backend in settings.AUTHENTICATION_BACKENDS:
        if not settings.DEBUG and issubclass(
                import_string(backend), MasterPasswordMixin):
            warnings.warn(PRODUCTION_WARNING)
            break


production_warning()



import getpass
import sys

from django.contrib.auth.hashers import make_password
from django.core.management.base import BaseCommand
from six.moves import input


class Command(BaseCommand):
    help = 'Creates a hashed version of a password.'

    def handle(self, **options):
        # We can't redirect stdin to `getpass()` so we have to fallback to
        # `input()` to test. See: http://pymotw.com/2/getpass/
        prompt = getpass.getpass if sys.stdin.isatty() else input
        try:
            password = make_password(prompt())
        # TODO: Not sure if there is any way to test this?
        except KeyboardInterrupt:
            self.stderr.write("\nOperation cancelled.")
            sys.exit(1)
        self.stdout.write('Hashed password: %s' % password)



"""
Tests for ``master_password`` app.
"""

# WebTest API docs: http://webtest.readthedocs.org/en/latest/api.html

import string
import warnings

from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password
from django.core.management import call_command
from django.test import override_settings
from django.test.utils import captured_stdin, captured_stdout
from django_dynamic_fixture import G
from django_webtest import WebTest

from master_password.auth import PRODUCTION_WARNING, production_warning

User = get_user_model()

CHARS = string.digits + string.ascii_lowercase + string.ascii_uppercase + \
    string.punctuation

MASTER_PASSWORDS = {
    'user123': lambda u: not u.is_staff and not u.is_superuser,
    make_password('staff123'): lambda u: u.is_staff,
    'superuser123': None,
    'strong' + CHARS: None,
    make_password('hashed' + CHARS): None
}


class Auth(WebTest):
    """
    Tests for the ``auth`` module.
    """

    @override_settings(DEBUG=True, MASTER_PASSWORDS=MASTER_PASSWORDS)
    def test_ModelBackend(self):
        # Create a user account.
        user = G(User)
        user.set_password(CHARS)
        user.save()

        # Create a staff account.
        staff = G(User, is_staff=True)
        staff.set_password(CHARS)
        staff.save()

        # Create a superuser account.
        superuser = G(User, is_superuser=True)
        superuser.set_password(CHARS)
        superuser.save()

        # Normal password validation still works.
        self.assertFalse(self.client.login(username='wrong', password='wrong'))
        self.assertFalse(self.client.login(username='wrong', password=CHARS))
        self.assertFalse(self.client.login(
            username=user.username, password='wrong'))
        self.assertTrue(self.client.login(
            username=user.username, password=CHARS))

        # Master passwords can apply to all users, if no callback is defined.
        self.assertTrue(self.client.login(
            username=user.username, password='superuser123'))
        self.assertTrue(self.client.login(
            username=superuser.username, password='superuser123'))

        # Or just a subset of users.
        self.assertTrue(self.client.login(
            username=user.username, password='user123'))
        self.assertFalse(self.client.login(
            username=superuser.username, password='user123'))

        # You can store hashed master passwords instead of clear text.
        self.assertTrue(self.client.login(
            username=staff.username, password='staff123'))

        # When DEBUG is False, you *must* use strong, hashed master passwords.
        with override_settings(DEBUG=False):
            # Weak, clear text.
            self.assertFalse(self.client.login(
                username=user.username, password='superuser123'))
            # Weak, hashed.
            self.assertFalse(self.client.login(
                username=staff.username, password='staff123'))
            # Strong, clear text.
            self.assertFalse(self.client.login(
                username=user.username, password='strong' + CHARS))
            # Strong, hashed.
            self.assertTrue(self.client.login(
                username=user.username, password='hashed' + CHARS))

    # @override_settings(DEBUG=False)
    # def test_production_warning(self):
    #     # See: https://docs.python.org/3/library/warnings.html#testing-warnings
    #     with warnings.catch_warnings(record=True) as w:
    #         warnings.simplefilter('always')
    #         production_warning()
    #         self.assertEqual(len(w), 1)
    #         self.assertEqual('%s' % w[-1].message, PRODUCTION_WARNING)


class Management(WebTest):
    """
    Tests for management commands.
    """

    def test_make_password(self):
        with captured_stdin() as stdin, captured_stdout() as stdout:
            stdin.write('password123')
            stdin.seek(0)
            call_command('make_password')
            self.assertIn(
                'Hashed password: pbkdf2_sha256$', stdout.getvalue())



from __future__ import print_function

import setuptools
import sys

# Convert README.md to reStructuredText.
if {'bdist_wheel', 'sdist'}.intersection(sys.argv):
    try:
        import pypandoc
    except ImportError:
        print('WARNING: You should install `pypandoc` to convert `README.md` '
              'to reStructuredText to use as long description.',
              file=sys.stderr)
    else:
        print('Converting `README.md` to reStructuredText to use as long '
              'description.')
        long_description = pypandoc.convert('README.md', 'rst')

setuptools.setup(
    name='chattr-django-master-password',
    version='1.1.2',
    author='Interaction Consortium',
    author_email='studio@interaction.net.au',
    url='https://github.com/ixc/django-master-password',
    description='Login as any user with a master password. Add master '
                'password support to your custom authentication backend.',
    long_description=locals().get('long_description', ''),
    license='MIT',
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'Django',
    ],
    extras_require={
        'dev': [
            'ipdb',
            'ipython',
        ],
        'test': [
            'coverage',
            'coveralls',
            'django-dynamic-fixture',
            'django-nose',
            'django-webtest',
            'nose-progressive',
            'WebTest',
        ],
    },
    setup_requires=['setuptools_scm'],
)



"""
App configuration for ``master_password`` app.
"""

# Register signal handlers, but avoid interacting with the database.
# See: https://docs.djangoproject.com/en/1.8/ref/applications/#django.apps.AppConfig.ready

from django.apps import AppConfig


class AppConfig(AppConfig):
    name = 'master_password'
